"""The aaargs library to help with attribute autocompletion and argparse library"""
import argparse
import importlib.metadata
import typing

import zninit

__version__ = importlib.metadata.version("aaargs")


class ArgumentParser(zninit.ZnInit):
    """Define a dataclass like container for the argparse library"""

    prog = None
    usage = None
    description = None
    epilog = None
    parents = None
    formatter_class = argparse.HelpFormatter
    prefix_chars = "-"
    fromfile_prefix_chars = None
    argument_default = None
    conflict_handler = "error"
    add_help = True
    allow_abbrev = True

    def __init_subclass__(cls, **kwargs):
        """Allow adding arguments through subclass creation"""
        super().__init_subclass__()
        for key in kwargs:
            if key in dir(cls):
                setattr(cls, key, kwargs[key])
            else:
                raise AttributeError(f"Class {cls} has no attribute '{key}'.")
        return cls

    @classmethod
    def get_parser(cls) -> argparse.ArgumentParser:
        """Get the ArgumentParser object based on class attributes."""
        if cls.parents is None:
            cls.parents = []
        parser = argparse.ArgumentParser(
            prog=cls.prog,
            usage=cls.usage,
            description=cls.description,
            epilog=cls.epilog,
            parents=cls.parents,
            formatter_class=cls.formatter_class,
            prefix_chars=cls.prefix_chars,
            fromfile_prefix_chars=cls.fromfile_prefix_chars,
            argument_default=cls.argument_default,
            conflict_handler=cls.conflict_handler,
            add_help=cls.add_help,
            allow_abbrev=cls.allow_abbrev,
        )
        arguments: typing.List[Argument] = cls._get_descriptors()

        for argument in arguments:
            parser.add_argument(*argument.name_or_flags, **argument.kwargs)

        return parser

    @classmethod
    def parse_args(cls, args=None, namespace=None):
        """Run parse_args from the argparse.ArgumentParser

        Parameters
        ----------
        args: List of strings to parse. The default is taken from sys.argv
        namespace: An object to take the attributes.

        Returns
        -------
        an instance of 'self' with all attributes set.

        """
        parser = cls.get_parser()
        args = parser.parse_args(args, namespace)
        try:
            return cls(**args.__dict__)
        except TypeError as err:
            arguments: typing.List[Argument] = cls._get_descriptors()
            argument_names = [
                argument.name
                for argument in arguments
                if argument.name not in args.__dict__
            ]
            raise AttributeError(
                f"Arguments '{argument_names}' not in '{args}'. Check that the"
                f" attribute names {argument_names} match the argparse names. E.g."
                " 'filename = Argument(--filename)'."
            ) from err


class Argument(zninit.Descriptor):
    """An argparse argument."""

    def __init__(
        self,
        *name_or_flags: str,
        action=None,
        choices=None,
        const=None,
        default=None,
        dest=None,
        help=None,
        metavar=None,
        nargs=None,
        required=None,
        type=None,
        positional=None,
    ):
        """Replace the argparse.ArgumentParser.add_argument method.

        All arguments are passed to argparse.ArgumentParser.add_argument

        Parameters
        ----------
        positional: bool, default = True. Make the argument positional or keyword
            argument, if no name_or_flags are provided.

        """
        if default is not None:
            super().__init__(default=default)
        else:
            super().__init__()
        self.name_or_flags = name_or_flags
        self.positional = positional

        self.kwargs = {}

        if action is not None:
            self.kwargs["action"] = action
        if choices is not None:
            self.kwargs["choices"] = choices
        if const is not None:
            self.kwargs["const"] = const
        if default is not None:
            self.kwargs["default"] = default
        if dest is not None:
            self.kwargs["dest"] = dest
        if help is not None:
            self.kwargs["help"] = help
        if metavar is not None:
            self.kwargs["metavar"] = metavar
        if nargs is not None:
            self.kwargs["nargs"] = nargs
        if required is not None:
            self.kwargs["required"] = required
        if type is not None:
            self.kwargs["type"] = type

    def __get__(self, instance, owner=None):
        """Get method of the descriptor

        This class is used to set the name and allows for the special case:

        >>> class MyArgs(ArgumentParser):
        >>>     filename = Argument()
        >>>     verbose: bool = Argument()

        which will define a positional argument without defining 'name_or_flags'.
        When using 'positional=False' it will be converted to a keyword only argument.
        Futhermore, it allows for boolean arguments without defining 'positional=False'
        or 'action=store_true' explicitly.

        """
        if (
            self.owner.__annotations__.get(self.name) in ["bool", bool]
            and self.kwargs.get("action") is None
        ):
            self.kwargs["action"] = "store_true"
            if len(self.name_or_flags) == 0:
                if self.positional:
                    raise TypeError(
                        "Can not use boolean annotation with positional only Argument"
                        f" '{self.name}'"
                    )
                if self.default not in (True, False, zninit.Empty):
                    raise ValueError(
                        f"Default value for boolean argument '{self.name}' can only be"
                        f" boolean, not '{self.default}'"
                    )
                self.name_or_flags = (f"--{self.name}",)

        if self.positional is None:
            self.positional = True

        if len(self.name_or_flags) == 0:
            self.name_or_flags = (self.name if self.positional else f"--{self.name}",)

        return super().__get__(instance, owner)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aaargs']

package_data = \
{'': ['*']}

install_requires = \
['zninit>=0.1.2,<0.2.0']

setup_kwargs = {
    'name': 'aaargs',
    'version': '0.1.1',
    'description': 'autocompletion and argument parsing',
    'long_description': '[![coveralls](https://coveralls.io/repos/github/zincware/aaargs/badge.svg)](https://coveralls.io/github/zincware/aaargs)\n![PyTest](https://github.com/zincware/aaargs/actions/workflows/pytest.yaml/badge.svg)\n[![PyPI version](https://badge.fury.io/py/aaargs.svg)](https://badge.fury.io/py/aaargs)\n\n# Aaargs ...\n\nI\'m not a huge fan of the [argparse](https://docs.python.org/3/library/argparse.html) library that ships with Python.\nPersonally, I much prefer  [typer](https://typer.tiangolo.com/) or [click](https://click.palletsprojects.com/).\nBut `argparse` is often used so this is my approach in bringing at least **a**ttribute **a**utocompletion to the **arg**parse library.\n\nLet us take a look at the official documentation and use their examples:\n\n```python\nimport argparse\n\nparser = argparse.ArgumentParser(\n                    prog = \'ProgramName\',\n                    description = \'What the program does\',\n                    epilog = \'Text at the bottom of help\')\n\nparser.add_argument(\'filename\')           # positional argument\nparser.add_argument(\'-c\', \'--count\')      # option that takes a value\nparser.add_argument(\'-v\', \'--verbose\',\n                    action=\'store_true\')  # on/off flag\n\nargs = parser.parse_args()\nprint(args.filename, args.count, args.verbose)\n```\n\nWhy isn\'t the `argparse.ArgumentParser` a container class, like a dataclass?\n\nSo my approach to *solve* this looks like this:\n\n```python\nfrom aaargs import ArgumentParser, Argument\n\nclass MyParser(ArgumentParser):\n    rog = "ProgramName"\n    description = "What the program does"\n    epilog = "Text at the bottom of help"\n\n    # You can define arguments directly\n    filename = Argument()  # positional argument\n    encoding = Argument(positional=False)  # keyword argument \'--encoding\'\n    \n    # or pass the \'name_or_flags\' argument\n    count = Argument("-c", "--count")\n    verbose = Argument("-v", "--verbose", action="store_true")\n    \n    # for boolean types annotations are also supported\n    debug: bool = Argument() # --debug with action="store_true"\n\nparser: argparse.ArgumentParser = MyParser.get_parser()\nargs: MyParser = MyParser.parse_args()\n```\n\nYou can also print the parser just like the original:\n```python\nargs = MyParser.parse_args(\n        ["README.md", "--encoding", "utf-8", "-c", "3", "--debug"]\n    )\n\nprint(args)\n>>> MyParser(count=\'3\', debug=True, encoding=\'utf-8\', filename=\'README.md\', verbose=False)\nprint(args.encoding) # this will autocomplete 🎉\n>>> "utf-8"\n```\n\nYou can also create a Parser using keyword arguments if you prefer (I don\'t):\n\n```python\nfrom aaargs import ArgumentParser\n\nclass MyParser(\n    ArgumentParser,\n    prog="ProgramName",\n    description="What the program does",\n    epilog="Text at the bottom of help",\n):\n    ...\n```\n',
    'author': 'zincwarecode',
    'author_email': 'zincwarecode@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
